import functools
from archaea.geometry.point3d import Point3d
from archaea.geometry.plane import Plane
from archaea.geometry.vector3d import Vector3d


# FIXME: Inherit later from base LineSegment object to cover common functionality
#  to prevent duplicated codes
class LineSegment:
    start: Point3d
    end: Point3d

    def __init__(self, start: Point3d, end: Point3d):
        self.start = start
        self.end = end

    def __eq__(self, other):
        return self.start == other.start and self.end == other.end

    @classmethod
    def from_point_and_vector(cls, start: Point3d, vector: Vector3d):
        end_point = start.move(vector)
        return cls(start, end_point)

    @functools.cached_property
    def length(self):
        return self.start.distance_to(self.end)

    @functools.cached_property
    def vector(self):
        return self.start.vector_to(self.end)

    def reverse(self):
        return LineSegment(self.end, self.start)

    def point_at(self, t: float) -> Point3d:
        return self.start + self.vector.scale(t)

    def parameter_at(self, point: Point3d):
        return self.start.vector_to(point).dot(self.vector) / self.vector.magnitude()**2

    def is_point_on_line(self, point: Point3d):
        return point == self.point_at(self.parameter_at(point))

    def is_point_on_segment(self, point: Point3d):
        parameter = self.parameter_at(point)
        return self.is_point_on_line(point) and 0 <= parameter <= 1

    def closest_point(self, point: Point3d):
        closest_parameter = self.closest_point_parameter(point)
        if closest_parameter <= 0:
            return self.start
        elif closest_parameter >= 1:
            return self.end
        else:
            return self.point_at(closest_parameter)

    def closest_point_parameter(self, point: Point3d):
        return self.vector.dot(self.start.vector_to(point)) / self.length**2

    def distance_to_point(self, point: Point3d):
        closest_point = self.closest_point(point)
        return closest_point.distance_to(point)

    def normal_on_plane(self, plane: Plane):
        return self.vector.cross_product(plane.normal).normalize()

    def distance_to_segment(self, segment):
        if self.is_intersects(segment):
            return 0
        distances = [
            self.distance_to_point(segment.start),
            self.distance_to_point(segment.end),
            segment.distance_to_point(self.start),
            segment.distance_to_point(self.end)
        ]
        return min(distances)

    # Thanks to https://stackoverflow.com/questions/55220355/how-to-detect-whether-two-segments-in-3d-space-intersect
    def is_intersects(self, segment):
        p1 = self.start.position_vector
        p2 = self.end.position_vector
        q1 = segment.start.position_vector
        q2 = segment.end.position_vector
        p_segment_vector = p2 - p1
        q_segment_vector = q2 - q1
        qp_distance_vector = q1 - p1
        p_length = self.length

        a = p_segment_vector.dot(qp_distance_vector) / p_length**2
        b = p_segment_vector.dot(q_segment_vector) / p_length**2
        c = p_segment_vector.scale(b) - q_segment_vector
        c_length = c.length
        if c_length == 0:
            c_length = 1
        t1 = c.dot(q1 - p1.scale(1 - a) - p2.scale(a)) / c_length**2
        t0 = a + (t1 * b)
        is_intersects = (0 <= t0 <= 1) and (0 <= t1 <= 1)
        return is_intersects

    def extrude(self, vector):
        start = self.start
        end = self.end
        extruded_start = start.move(vector)
        extruded_end = end.move(vector)
        from archaea.geometry.loop import Loop
        from archaea.geometry.face import Face
        loop = Loop([start, end, extruded_end, extruded_start])
        return Face(loop)

    def move(self, vector):
        moved_start = self.start.move(vector)
        moved_end = self.end.move(vector)
        return LineSegment(moved_start, moved_end)

    @staticmethod
    def line_intersection(line1, line2):
        plane = Plane.from_3_point(line1.start, line1.end, line2.start)
        line1_uv = [plane.plane_coordinates(line1.start), plane.plane_coordinates(line1.end)]
        line2_uv = [plane.plane_coordinates(line2.start), plane.plane_coordinates(line2.end)]
        intersected_uv = LineSegment.line_intersection_uv(line1_uv, line2_uv)
        return plane.point_at(intersected_uv[0], intersected_uv[1])

    @staticmethod
    def line_intersection_uv(line1, line2):
        xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
        ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

        def det(a, b):
            return a[0] * b[1] - a[1] * b[0]

        div = det(xdiff, ydiff)
        if div == 0:
            raise Exception('lines do not intersect')
        d = (det(*line1), det(*line2))
        x = det(d, xdiff) / div
        y = det(d, ydiff) / div
        return x, y

    @staticmethod
    def line_intersection_2(line1, line2):
        xdiff = (line1.start.x - line1.end.x, line2.start.x - line2.end.x)
        ydiff = (line1.start.y - line1.end.y, line2.start.y - line2.end.y)

        def det(a, b):
            return a[0] * b[1] - a[1] * b[0]

        div = det(xdiff, ydiff)
        if div == 0:
            raise Exception('lines do not intersect')
        d = (det([line1.start.x, line1.start.y], [line1.end.x, line1.end.y]),
             det([line2.start.x, line2.start.y], [line2.end.x, line2.end.y]))
        x = det(d, xdiff) / div
        y = det(d, ydiff) / div
        return x, y



# Dimensionless coordinate array for integer and float values.
class CoordinateArray(list):
    coordinates: "list[(int,float)]"

    def __init__(self, *coordinates):
        filtered_coordinates = [entry for entry in coordinates if isinstance(entry, (int, float))]
        self.coordinates = filtered_coordinates
        super().__init__(filtered_coordinates)

    @property
    def dimension(self):
        return len(self)

    def __add__(self, other):
        return self.add(other)

    def __sub__(self, other):
        return self.add(other.scale(-1))

    def __eq__(self, other):
        if self.dimension != other.dimension:
            return False
        return (self - other).magnitude() == 0

    def dot(self, other):
        if self.dimension == 2:
            return (self[0] * other[0]) + (self[1] * other[1])
        elif self.dimension == 3:
            return (self[0] * other[0]) + (self[1] * other[1]) + (self[2] * other[2])
        else:
            sum([value_1 + value_2 for (value_1, value_2) in zip(self, other)])

    def magnitude(self):
        return (self.dot(self)) ** 0.5

    def cos(self, other):
        return self.dot(other) / (self.magnitude() * other.magnitude())

    def sin(self, other):
        return (1 - (self.cos(other) ** 2)) ** 0.5

    def add(self, vector_2):
        summed_coordinates = [value_1 + value_2 for (value_1, value_2) in zip(self, vector_2)]
        return self.__class__(*summed_coordinates)

    def scale(self, scale):
        scaled_coordinates = [coordinate * scale for coordinate in self.coordinates]
        return self.__class__(*scaled_coordinates)



import math

__all__ = ['earcut', 'deviation', 'flatten']


def earcut(data, hole_indices=None, dim=None):
    dim = dim or 2

    has_holes = hole_indices and len(hole_indices)
    outer_len = hole_indices[0] * dim if has_holes else len(data)
    outer_node = linked_list(data, 0, outer_len, dim, True)
    triangles = []

    if not outer_node:
        return triangles

    min_x = None
    min_y = None
    max_x = None
    max_y = None
    x = None
    y = None
    size = None

    if has_holes:
        outer_node = eliminate_holes(data, hole_indices, outer_node, dim)

    # if the shape is not too simple, we'll use z-order curve hash later; calculate polygon bbox
    if len(data) > 80 * dim:
        min_x = max_x = data[0]
        min_y = max_y = data[1]

        for i in range(dim, outer_len, dim):
            x = data[i]
            y = data[i + 1]
            if x < min_x:
                min_x = x
            if y < min_y:
                min_y = y
            if x > max_x:
                max_x = x
            if y > max_y:
                max_y = y

        # minX, minY and size are later used to transform coords into integers for z-order calculation
        size = max(max_x - min_x, max_y - min_y)

    earcut_linked(outer_node, triangles, dim, min_x, min_y, size)

    return triangles


# create a circular doubly linked _list from polygon points in the specified winding order
def linked_list(data, start, end, dim, clockwise):
    i = None
    last = None

    if clockwise == (signed_area(data, start, end, dim) > 0):
        for i in range(start, end, dim):
            last = insert_node(i, data[i], data[i + 1], last)

    else:
        for i in reversed(range(start, end, dim)):
            last = insert_node(i, data[i], data[i + 1], last)

    if last and equals(last, last.next):
        remove_node(last)
        last = last.next

    return last


# eliminate collinear or duplicate points
def filter_points(start, end=None):
    if not start:
        return start
    if not end:
        end = start

    p = start
    again = True

    while again or p != end:
        again = False

        if not p.steiner and (equals(p, p.next) or area(p.prev, p, p.next) == 0):
            remove_node(p)
            p = end = p.prev
            if p == p.next:
                return None

            again = True

        else:
            p = p.next

    return end


# main ear slicing loop which triangulates a polygon (given as a linked _list)
def earcut_linked(ear, triangles, dim, minX, minY, size, _pass=None):
    if not ear:
        return

    # interlink polygon nodes in z-order
    if not _pass and size:
        index_curve(ear, minX, minY, size)

    stop = ear
    prev = None
    next = None

    # iterate through ears, slicing them one by one
    while ear.prev != ear.next:
        prev = ear.prev
        next = ear.next

        if is_ear_hashed(ear, minX, minY, size) if size else is_ear(ear):
            # cut off the triangle
            triangles.append(prev.i // dim)
            triangles.append(ear.i // dim)
            triangles.append(next.i // dim)

            remove_node(ear)

            # skipping the next vertex leads to less sliver triangles
            ear = next.next
            stop = next.next

            continue

        ear = next

        # if we looped through the whole remaining polygon and can't find any more ears
        if ear == stop:
            # try filtering points and slicing again
            if not _pass:
                earcut_linked(filter_points(ear), triangles, dim, minX, minY, size, 1)

                # if this didn't work, try curing all small self-intersections locally
            elif _pass == 1:
                ear = cure_local_intersections(ear, triangles, dim)
                earcut_linked(ear, triangles, dim, minX, minY, size, 2)

                # as a last resort, try splitting the remaining polygon into two
            elif _pass == 2:
                split_earcut(ear, triangles, dim, minX, minY, size)

            break


# check whether a polygon node forms a valid ear with adjacent nodes
def is_ear(ear):
    a = ear.prev
    b = ear
    c = ear.next

    if area(a, b, c) >= 0:
        return False  # reflex, can't be an ear

    # now make sure we don't have other points inside the potential ear
    p = ear.next.next

    while p != ear.prev:
        if point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) and area(p.prev, p, p.next) >= 0:
            return False
        p = p.next

    return True


def is_ear_hashed(ear, min_x, min_y, size):
    a = ear.prev
    b = ear
    c = ear.next

    if area(a, b, c) >= 0:
        return False # reflex, can't be an ear

    # triangle bbox; min & max are calculated like this for speed
    min_tx = (a.x if a.x < c.x else c.x) if a.x < b.x else (b.x if b.x < c.x else c.x)
    min_ty = (a.y if a.y < c.y else c.y) if a.y < b.y else (b.y if b.y < c.y else c.y)
    max_tx = (a.x if a.x > c.x else c.x) if a.x > b.x else (b.x if b.x > c.x else c.x)
    max_ty = (a.y if a.y > c.y else c.y) if a.y > b.y else (b.y if b.y > c.y else c.y)

    # z-order range for the current triangle bbox;
    min_z = z_order(min_tx, min_ty, min_x, min_y, size)
    max_z = z_order(max_tx, max_ty, min_x, min_y, size)

    # first look for points inside the triangle in increasing z-order
    p = ear.nextZ

    while p and p.z <= max_z:
        if p != ear.prev and p != ear.next and point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) and area(p.prev, p, p.next) >= 0:
            return False
        p = p.nextZ

    # then look for points in decreasing z-order
    p = ear.prevZ

    while p and p.z >= min_z:
        if p != ear.prev and p != ear.next and point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) and area(p.prev, p, p.next) >= 0:
            return False
        p = p.prevZ

    return True


# go through all polygon nodes and cure small local self-intersections
def cure_local_intersections(start, triangles, dim):
    do = True
    p = start

    while do or p != start:
        do = False

        a = p.prev
        b = p.next.next

        if not equals(a, b) and intersects(a, p, p.next, b) and locally_inside(a, b) and locally_inside(b, a):
            triangles.append(a.i // dim)
            triangles.append(p.i // dim)
            triangles.append(b.i // dim)

            # remove two nodes involved
            remove_node(p)
            remove_node(p.next)

            p = start = b

        p = p.next

    return p


# try splitting polygon into two and triangulate them independently
def split_earcut(start, triangles, dim, min_x, min_y, size):
    # look for a valid diagonal that divides the polygon into two
    do = True
    a = start

    while do or a != start:
        do = False
        b = a.next.next

        while b != a.prev:
            if a.i != b.i and is_valid_diagonal(a, b):
                # split the polygon in two by the diagonal
                c = split_polygon(a, b)

                # filter collinear points around the cuts
                a = filter_points(a, a.next)
                c = filter_points(c, c.next)

                # run earcut on each half
                earcut_linked(a, triangles, dim, min_x, min_y, size)
                earcut_linked(c, triangles, dim, min_x, min_y, size)
                return

            b = b.next

        a = a.next


# link every hole into the outer loop, producing a single-ring polygon without holes
def eliminate_holes(data, hole_indices, outer_node, dim):
    queue = []
    i = None
    _len = len(hole_indices)
    start = None
    end = None
    _list = None

    for i in range(len(hole_indices)):
        start = hole_indices[i] * dim
        end = hole_indices[i + 1] * dim if i < _len - 1 else len(data)
        _list = linked_list(data, start, end, dim, False)

        if _list == _list.next:
            _list.steiner = True

        queue.append(get_leftmost(_list))

    queue = sorted(queue, key=lambda i: i.x)

    # process holes from left to right
    for i in range(len(queue)):
        eliminate_hole(queue[i], outer_node)
        outer_node = filter_points(outer_node, outer_node.next)

    return outer_node


# find a bridge between vertices that connects hole with an outer ring and and link it
def eliminate_hole(hole, outer_node):
    outer_node = find_hole_bridge(hole, outer_node)
    if outer_node:
        b = split_polygon(outer_node, hole)
        filter_points(b, b.next)


# David Eberly's algorithm for finding a bridge between hole and outer polygon
def find_hole_bridge(hole, outer_node):
    do = True
    p = outer_node
    hx = hole.x
    hy = hole.y
    qx = -math.inf
    m = None

    # find a segment intersected by a ray from the hole's leftmost point to the left;
    # segment's endpoint with lesser x will be potential connection point
    while do or p != outer_node:
        do = False
        if p.y >= hy >= p.next.y and p.next.y - p.y != 0:
            x = p.x + (hy - p.y) * (p.next.x - p.x) / (p.next.y - p.y)

            if hx >= x > qx:
                qx = x

                if x == hx:
                    if hy == p.y:
                        return p
                    if hy == p.next.y:
                        return p.next

                m = p if p.x < p.next.x else p.next

        p = p.next

    if not m:
        return None

    if hx == qx:
        return m.prev # hole touches outer segment; pick lower endpoint

    # look for points inside the triangle of hole point, segment intersection and endpoint;
    # if there are no points found, we have a valid connection;
    # otherwise choose the point of the minimum angle with the ray as connection point

    stop = m
    mx = m.x
    my = m.y
    tan_min = math.inf
    tan = None

    p = m.next

    while p != stop:
        hx_or_qx = hx if hy < my else qx
        qx_or_hx = qx if hy < my else hx

        if hx >= p.x >= mx and point_in_triangle(hx_or_qx, hy, mx, my, qx_or_hx, hy, p.x, p.y):

            tan = abs(hy - p.y) / (hx - p.x)  # tangential

            if (tan < tan_min or (tan == tan_min and p.x > m.x)) and locally_inside(p, hole):
                m = p
                tan_min = tan

        p = p.next

    return m


# interlink polygon nodes in z-order
def index_curve(start, min_x, min_y, size):
    do = True
    p = start

    while do or p != start:
        do = False

        if p.z is None:
            p.z = z_order(p.x, p.y, min_x, min_y, size)

        p.prevZ = p.prev
        p.nextZ = p.next
        p = p.next

    p.prevZ.nextZ = None
    p.prevZ = None

    sort_linked(p)


# Simon Tatham's linked _list merge sort algorithm
# http:#www.chiark.greenend.org.uk/~sgtatham/algorithms/_listsort.html
def sort_linked(_list):
    do = True
    i = None
    p = None
    q = None
    e = None
    tail = None
    num_merges = None
    p_size = None
    q_size = None
    in_size = 1

    while do or num_merges > 1:
        do = False
        p = _list
        _list = None
        tail = None
        num_merges = 0

        while p:
            num_merges += 1
            q = p
            p_size = 0
            for i in range(in_size):
                p_size += 1
                q = q.nextZ
                if not q:
                    break

            q_size = in_size

            while p_size > 0 or (q_size > 0 and q):

                if p_size == 0:
                    e = q
                    q = q.nextZ
                    q_size -= 1

                elif q_size == 0 or not q:
                    e = p
                    p = p.nextZ
                    p_size -= 1

                elif p.z <= q.z:
                    e = p
                    p = p.nextZ
                    p_size -= 1

                else:
                    e = q
                    q = q.nextZ
                    q_size -= 1

                if tail:
                    tail.nextZ = e

                else:
                    _list = e

                e.prevZ = tail
                tail = e

            p = q

        tail.nextZ = None
        in_size *= 2

    return _list


# z-order of a point given coords and size of the data bounding box
def z_order(x, y, min_x, min_y, size):
    # coords are transformed into non-negative 15-bit integer range
    x = 32767 * (x - min_x) // size
    y = 32767 * (y - min_y) // size

    x = (x | (x << 8)) & 0x00FF00FF
    x = (x | (x << 4)) & 0x0F0F0F0F
    x = (x | (x << 2)) & 0x33333333
    x = (x | (x << 1)) & 0x55555555

    y = (y | (y << 8)) & 0x00FF00FF
    y = (y | (y << 4)) & 0x0F0F0F0F
    y = (y | (y << 2)) & 0x33333333
    y = (y | (y << 1)) & 0x55555555

    return x | (y << 1)


# find the leftmost node of a polygon ring
def get_leftmost(start):
    do = True
    p = start
    leftmost = start

    while do or p != start:
        do = False
        if p.x < leftmost.x:
            leftmost = p
        p = p.next

    return leftmost


# check if a point lies within a convex triangle
def point_in_triangle(ax, ay, bx, by, cx, cy, px, py):
    return (cx - px) * (ay - py) - (ax - px) * (cy - py) >= 0 and \
        (ax - px) * (by - py) - (bx - px) * (ay - py) >= 0 and \
        (bx - px) * (cy - py) - (cx - px) * (by - py) >= 0


# check if a diagonal between two polygon nodes is valid (lies in polygon interior)
def is_valid_diagonal(a, b):
    return a.next.i != b.i and a.prev.i != b.i and not intersects_polygon(a, b) and \
        locally_inside(a, b) and locally_inside(b, a) and middle_inside(a, b)


# signed area of a triangle
def area(p, q, r):
    return (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y)


# check if two points are equal
def equals(p1, p2):
    return p1.x == p2.x and p1.y == p2.y


# check if two segments intersect
def intersects(p1, q1, p2, q2):
    if (equals(p1, q1) and equals(p2, q2)) or (equals(p1, q2) and equals(p2, q1)):
        return True

    return area(p1, q1, p2) > 0 != area(p1, q1, q2) > 0 and \
        area(p2, q2, p1) > 0 != area(p2, q2, q1) > 0


# check if a polygon diagonal intersects any polygon segments
def intersects_polygon(a, b):
    do = True
    p = a

    while do or p != a:
        do = False
        if p.i != a.i and p.next.i != a.i and p.i != b.i and p.next.i != b.i and intersects(p, p.next, a, b):
            return True

        p = p.next

    return False


# check if a polygon diagonal is locally inside the polygon
def locally_inside(a, b):
    if area(a.prev, a, a.next) < 0:
        return area(a, b, a.next) >= 0 and area(a, a.prev, b) >= 0
    else:
        return area(a, b, a.prev) < 0 or area(a, a.next, b) < 0


# check if the middle point of a polygon diagonal is inside the polygon
def middle_inside(a, b):
    do = True
    p = a
    inside = False
    px = (a.x + b.x) / 2
    py = (a.y + b.y) / 2

    while do or p != a:
        do = False
        if ((p.y > py) != (p.next.y > py)) and (px < (p.next.x - p.x) * (py - p.y) / (p.next.y - p.y) + p.x):
            inside = not inside

        p = p.next

    return inside


# link two polygon vertices with a bridge; if the vertices belong to the same ring, it splits polygon into two;
# if one belongs to the outer ring and another to a hole, it merges it into a single ring
def split_polygon(a, b):
    a2 = Node(a.i, a.x, a.y)
    b2 = Node(b.i, b.x, b.y)
    an = a.next
    bp = b.prev

    a.next = b
    b.prev = a

    a2.next = an
    an.prev = a2

    b2.next = a2
    a2.prev = b2

    bp.next = b2
    b2.prev = bp

    return b2


# create a node and optionally link it with previous one (in a circular doubly linked _list)
def insert_node(i, x, y, last):
    p = Node(i, x, y)

    if not last:
        p.prev = p
        p.next = p

    else:
        p.next = last.next
        p.prev = last
        last.next.prev = p
        last.next = p

    return p


def remove_node(p):
    p.next.prev = p.prev
    p.prev.next = p.next

    if p.prevZ:
        p.prevZ.nextZ = p.nextZ

    if p.nextZ:
        p.nextZ.prevZ = p.prevZ


class Node(object):
    def __init__(self, i, x, y):
        # vertex index in coordinates array
        self.i = i

        # vertex coordinates

        self.x = x
        self.y = y

        # previous and next vertex nodes in a polygon ring
        self.prev = None
        self.next = None

        # z-order curve value
        self.z = None

        # previous and next nodes in z-order
        self.prevZ = None
        self.nextZ = None

        # indicates whether this is a steiner point
        self.steiner = False


# return a percentage difference between the polygon area and its triangulation area;
# used to verify correctness of triangulation
def deviation(data, hole_indices, dim, triangles):
    _len = len(hole_indices)
    has_holes = hole_indices and len(hole_indices)
    outer_len = hole_indices[0] * dim if has_holes else len(data)

    polygon_area = abs(signed_area(data, 0, outer_len, dim))

    if has_holes:
        for i in range(_len):
            start = hole_indices[i] * dim
            end = hole_indices[i + 1] * dim if i < _len - 1 else len(data)
            polygon_area -= abs(signed_area(data, start, end, dim))

    triangles_area = 0

    for i in range(0, len(triangles), 3):
        a = triangles[i] * dim
        b = triangles[i + 1] * dim
        c = triangles[i + 2] * dim
        triangles_area += abs(
            (data[a] - data[c]) * (data[b + 1] - data[a + 1]) -
            (data[a] - data[b]) * (data[c + 1] - data[a + 1]))

    if polygon_area == 0 and triangles_area == 0:
        return 0

    return abs((triangles_area - polygon_area) / polygon_area)


def signed_area(data, start, end, dim):
    sum_area = 0
    j = end - dim

    for i in range(start, end, dim):
        sum_area += (data[j] - data[i]) * (data[i + 1] + data[j + 1])
        j = i

    return sum_area


# turn a polygon in a multi-dimensional array form (e.g. as in GeoJSON) into a form Earcut accepts
def flatten(data):
    dim = len(data[0][0])
    result = {
        'vertices': [],
        'holes': [],
        'dimensions': dim
    }
    hole_index = 0

    for i in range(len(data)):
        for j in range(len(data[i])):
            for d in range(dim):
                result['vertices'].append(data[i][j][d])

        if i > 0:
            hole_index += len(data[i - 1])
            result['holes'].append(hole_index)

    return result


def unflatten(data):
    result = []

    for i in range(0, len(data), 3):
        result.append(tuple(data[i:i + 3]))

    return result



import functools
from archaea.geometry.line_segment import LineSegment
from archaea.geometry.point3d import Point3d


class Polyline:
    points: "list[Point3d]"

    def __init__(self, points: "list[Point3d]"):
        self.points = points

    @functools.cached_property
    def segments(self):
        consecutive_pairs = [[self.points[i], self.points[i + 1]]
                             for i in range(len(self.points) - 1)]
        return [LineSegment(start, end) for [start, end] in consecutive_pairs]

    @functools.cached_property
    def segment_count(self):
        return len(self.segments)

    @functools.cached_property
    def length(self):
        return sum(segment.length for segment in self.segments)

    @functools.cached_property
    def point_at_start(self) -> Point3d:
        return self.segments[0].start

    @functools.cached_property
    def point_at_end(self) -> Point3d:
        return self.segments[-1].end

    @functools.cached_property
    def is_closed(self):
        return self.point_at_start == self.point_at_end




import functools
from archaea.geometry.vector3d import Vector3d
from archaea.geometry.point3d import Point3d


class Plane:
    u: Vector3d
    v: Vector3d
    origin: Point3d

    def __init__(self, origin, u, v):
        if u.dot(v) > 1e-5:
            raise 'u-v should be orthogonal'
        self.u = u
        self.v = v
        self.origin = origin

    @functools.cached_property
    def normal(self):
        return self.u.cross_product(self.v).normalize()

    @classmethod
    def from_3_point(cls, origin_point: Point3d, point_2: Point3d, point_3: Point3d):
        u_direction = origin_point.vector_to(point_2).normalize()
        normal = u_direction.cross_product(origin_point.vector_to(point_3))
        v_direction = u_direction.cross_product(normal.normalize())
        return cls(origin_point, u_direction, v_direction)

    def point_at(self, u, v):
        return self.origin.move(self.u.scale(u) + self.v.scale(v))

    def plane_coordinates(self, point):
        origin_to_point = self.origin.vector_to(point)
        coordinate_u = origin_to_point.dot(self.u)
        coordinate_v = origin_to_point.dot(self.v)
        return [coordinate_u, coordinate_v]

    def project_to_plane(self, point):
        plane_coordinates = self.plane_coordinates(point)
        return self.point_at(plane_coordinates[0], plane_coordinates[1])

    def project_to_plane_on_z(self, point):
        d = self.normal.dot(self.origin.position_vector)
        z_value_on_plane = (d - (self.normal.x * point.x) - (self.normal.y * point.y)) / self.normal.z
        return Point3d(point.x, point.y, z_value_on_plane)

    def is_on_plane(self, point):
        return point.distance_to(self.project_to_plane(point)) < 1e-5

    def is_same_plane(self, other):
        return (self.is_on_plane(other.origin)) and self.is_on_plane(other.point_at(1, 0) and
                                                                     self.is_on_plane(other.point_at(0, 1)))

    def rotate(self, axis, angle, origin_point=None):
        if origin_point is None:
            origin_point = self.origin

        # Calculate the vector from the origin to the point
        vector_to_point = self.origin.vector_to(origin_point.position_vector)

        # Perform the rotation on the origin
        rotated_origin = origin_point.rotate(axis, angle)

        # Calculate the final position of the origin after rotation
        self.origin = rotated_origin

        # Rotate the u and v vectors
        self.u = self.u.rotate(axis, angle)
        self.v = self.v.rotate(axis, angle)



from archaea.writer.to_stl import to_stl
from archaea.earcut.earcut import *
from archaea.geometry.coordinate_array import CoordinateArray
from archaea.geometry.vector import Vector
from archaea.geometry.vector2d import Vector2d
from archaea.geometry.vector3d import Vector3d
from archaea.geometry.point2d import Point2d
from archaea.geometry.point3d import Point3d
from archaea.geometry.plane import Plane
from archaea.geometry.loop import Loop
from archaea.geometry.line_segment import LineSegment
from archaea.geometry.polyline import Polyline
from archaea.geometry.orientation import Orientation
from archaea.geometry.utils import *
from archaea.geometry.face import Face
from archaea.geometry.mesh import Mesh



import numpy as np
from stl import mesh
import stl

# faces = np.array(
#     [
#         [0, 3, 1],
#         [1, 3, 2],
#
#         [4, 7, 5],
#         [5, 7, 6],
#     ]
# )
#
# vertices = np.array(
#     [
#         [0, 0, 0], # 0
#         [10, 0, 0], # 1
#         [10, 10, 0], # 2
#         [0, 10, 0], # 3
#
#         [0, 0, 0],
#         [10, 0, 0],
#         [10, 0, 10],
#         [0, 0, 10]
#     ]
# )
#
# to_stl(faces, vertices, "test")


def to_stl(faces, vertices, file_path):
    geo = mesh.Mesh(np.zeros(faces.shape[0], dtype=mesh.Mesh.dtype))
    for i, f in enumerate(faces):
        for j in range(3):
            geo.vectors[i][j] = vertices[f[j], :]

    # Write the mesh to file
    geo.save('%s.stl' % file_path, mode=stl.Mode.ASCII)



import functools
from archaea.geometry.point3d import Point3d
from archaea.geometry.point2d import Point2d
from archaea.geometry.vector3d import Vector3d
from archaea.geometry.vector2d import Vector2d
from archaea.geometry.line_segment import LineSegment
from archaea.geometry.polyline import Polyline
from archaea.geometry.plane import Plane
from archaea.geometry.utils import area
from archaea.geometry.orientation import Orientation


class Loop(Polyline):
    points: "list[Point3d]"
    normal: Vector3d

    def __init__(self, points: "list[Point3d]"):
        if points[-1] != points[0]:
            points.append(points[0])
        self.points = points
        super().__init__(self.points)
        self.normal = self._calculate_normal(self.points)

    @functools.cached_property
    def segments(self):
        consecutive_pairs = [[self.points[i], self.points[i + 1]]
                             for i in range(len(self.points) - 1)]
        return [LineSegment(start, end) for [start, end] in consecutive_pairs]

    @functools.cached_property
    def area(self):
        poly = [point.to_a() for point in self.points[:-1]]
        return area(poly)

    @functools.cached_property
    def segment_count(self):
        return len(self.segments)

    # @functools.cached_property
    # def orientation(self) -> Orientation:
    #     if self.area == 0:
    #         return Orientation.UNDEFINED
    #     elif self.area > 0:
    #         return Orientation.COUNTER_CLOCKWISE
    #     else:
    #         return Orientation.CLOCKWISE

    def uv_points(self, plane=None) -> "list[Point2d]":
        plane = plane or self.plane()
        plane_coordinates = [plane.plane_coordinates(point) for point in self.points[:-1]]
        return [Point2d(pc[0], pc[1]) for pc in plane_coordinates]

    def reverse(self):
        points: "list[Point3d]" = list(reversed(self.points))
        return Loop(points)

    def point_from_uv(self, uv_point: Point2d) -> Point3d:
        plane = self.plane()
        return plane.point_at(uv_point.x, uv_point.y)

    def plane(self):
        return Plane.from_3_point(self.points[0], self.points[1], self.points[2])

    def to_face(self):
        from archaea.geometry.face import Face
        return Face(self)

    def move(self, vector):
        moved_points = []
        for point in self.points:
            moved_points.append(point.move(vector))
        return Loop(moved_points)

    def extrude(self, value, is_hole_loop=False):
        faces = []
        segments = self.segments if is_hole_loop else self.reverse().segments
        move_vector = Vector3d(*self.normal.scale(value))
        for line in segments:
            border = line.extrude(move_vector)
            faces.append(border)
        return faces

    def offset(self, value):
        # if self.orientation == Orientation.UNDEFINED:
        #     return self

        loop_plane = self.plane()
        sc = self.segment_count
        consecutive_segments = []
        for index, segment in enumerate(self.segments):
            consecutive_segments.append([segment, self.segments[(index + 1) % sc]])

        offset_points = []
        for pair in consecutive_segments:
            segment: LineSegment = pair[0]
            next_segment: LineSegment = pair[1]

            segment_normal = segment.normal_on_plane(loop_plane)
            next_normal = next_segment.normal_on_plane(loop_plane)

            moved_segment = segment.move(segment_normal.scale(value))
            moved_next_segment = next_segment.move(next_normal.scale(value))

            intersected_point = LineSegment.line_intersection(moved_segment, moved_next_segment)
            offset_points.append(intersected_point)

        return Loop(offset_points)

    def is_point_in_loop(self, point: Point3d) -> bool:
        plane = self.plane()
        on_plane = plane.is_on_plane(point)
        if not on_plane:
            return False

        # convert points to u,v coordinates
        plane_coordinates = [plane.plane_coordinates(point) for point in self.points]
        point_plane_coordinates = plane.plane_coordinates(point)

        sum_segment_length = sum([segment.length for segment in self.segments])
        sum_segment_distances = sum([segment.distance_to_point(point) for segment in self.segments])
        max_length = sum_segment_length + sum_segment_distances

        consecutive_pairs = [[plane_coordinates[i], plane_coordinates[i + 1]]
                             for i in range(len(plane_coordinates) - 1)]
        uv_segments = [LineSegment(Point3d(start[0], start[1], 0), Point3d(end[0], end[1], 0)) for [start, end] in consecutive_pairs]

        uv_point = Point3d(point_plane_coordinates[0], point_plane_coordinates[1], 0)
        # Create a point at infinity, y is same as point p
        x_line = LineSegment(uv_point, Point3d(max_length, point_plane_coordinates[1], 0))
        count = 0

        for uv_segment in uv_segments:
            if uv_segment.is_point_on_segment(uv_point):
                return True

            if uv_segment.is_intersects(x_line):
                count += 1

        #i = 0
        #n = len(plane_coordinates) - 1
        #while True:
        #    # Forming a line from two consecutive points of
        #    # poly
        #    pc = plane_coordinates[i]
        #    pc_next = plane_coordinates[i + 1]
        #    side = LineSegment(Point3d(pc[0], pc[1], 0), Point3d(pc_next[0], pc_next[1], 0))
        #    if x_line.is_intersects(side):
        #        # If side is intersects x_line
        #        if side.is_point_on_line(uv_point):
        #            return True
        #    count += 1
        #    i = (i + 1) % n
        #    if i != 0:
        #        break

        # True when count is odd
        return count % 2 == 1


    @staticmethod
    def _calculate_normal(points: "list[Point3d]") -> Vector3d:
        """
        Method that calculates normal from first point and vectors with second and third point.
        :param: points to calculate normal
        :returns normal as vector
        """
        v1: Vector3d = points[0].vector_to(points[1])
        v2: Vector3d = points[0].vector_to(points[2])
        return v1.cross_product(v2).normalize()



from typing import List

from archaea.geometry.point3d import Point3d
from archaea.geometry.face import Face
from archaea.writer.to_stl import to_stl
import numpy as np
import os


class Mesh:
    polygons: "list[list[int]]"
    vertices: "list[Point3d]"

    def __init__(self, polygons=None, vertices=None):
        if polygons is None:
            self.polygons = []
        else:
            self.polygons = polygons
        if vertices is None:
            self.vertices = []
        else:
            self.vertices = vertices

    def add_from_faces(self, faces: "list[Face]", share_vertices: bool = True):
        for face in faces:
            self.add_from_face(face, share_vertices)

    def add_from_face(self, face: Face, share_vertices: bool = True):
        face_vertices = face.mesh_polygon_vertices()
        for vertices in face_vertices:
            self.add_polygon(vertices, share_vertices)

    def add_polygon(self, vertices: "list[Point3d]", share_vertices: bool = True):
        polygon_indexes = []
        for vertex in vertices:
            if share_vertices:
                matches = [existing_vertex for existing_vertex in self.vertices if existing_vertex == vertex]
                if any(matches):
                    existing_vertex_index = self.vertices.index(matches[0])
                    polygon_indexes.append(existing_vertex_index)
                else:
                    self.vertices.append(vertex)
                    polygon_indexes.append(len(self.vertices) - 1)
            else:
                self.vertices.append(vertex)
                polygon_indexes.append(len(self.vertices) - 1)

        self.polygons.append(polygon_indexes)

    def to_stl(self, path, file_name):
        vertices = np.array([vertex.to_a() for vertex in self.vertices])
        faces = np.array(self.polygons)
        file_path = os.path.join(path, file_name)
        to_stl(faces, vertices, file_path)

    def find_average_point(self) -> Point3d:
        # Initialize variables to store the sum of coordinates
        sum_x = sum_y = sum_z = 0

        # Iterate through the list of points and sum up their coordinates
        for point in self.vertices:
            sum_x += point.x
            sum_y += point.y
            sum_z += point.z

        # Calculate the average by dividing by the number of points
        num_points = len(self.vertices)
        average_x = sum_x / num_points
        average_y = sum_y / num_points
        average_z = sum_z / num_points

        # Create and return the average point
        return Point3d(average_x, average_y, average_z)

    def rotate(self, axis, angle, origin=None):
        origin_to_rotate_around = origin
        if origin is None:
            origin_to_rotate_around = self.find_average_point()

        rotated_vertices = []
        for point in self.vertices:
            rotated_vertices.append(point.rotate(axis, angle, origin_to_rotate_around))

        return Mesh(self.polygons, rotated_vertices)



from archaea.geometry.coordinate_array import CoordinateArray
from archaea.geometry.vector2d import Vector2d


# 2 dimensional point.
class Point2d(CoordinateArray):
    def __init__(self, *coordinates):
        filtered_coordinates = [entry for entry in coordinates if isinstance(entry, (int, float))]
        self.coordinates = filtered_coordinates[:2]
        super(CoordinateArray, self).__init__(filtered_coordinates)

    @property
    def x(self):
        return self.coordinates[0]

    @property
    def y(self):
        return self.coordinates[1]

    def position_vector(self):
        return Vector2d(self.x, self.y)

    def move(self, vector):
        return Point2d(self.x + vector.x, self.y + vector.y)

    def vector_to(self, other):
        return Vector2d(other.x - self.x, other.y - self.y)

    def distance_to(self, other):
        return (((self.x - other.x) ** 2) + ((self.y - other.y) ** 2)) ** 0.5



from typing import TextIO
from pathlib import Path

from archaea.geometry.face import Face
from archaea.geometry.point3d import Point3d
from archaea.geometry.triangle import Triangle

p1 = Point3d(0, 0, 0)
p2 = Point3d(10, 0, 0)
p3 = Point3d(10, 10, 0)
p4 = Point3d(0, 10, 0)

p5 = Point3d(0, 0, 10)
p6 = Point3d(10, 0, 10)
p7 = Point3d(10, 10, 10)
p8 = Point3d(0, 10, 10)

triangle_floor_1 = Triangle([p1, p3, p2, p1])
triangle_floor_2 = Triangle([p1, p4, p3, p1])

triangle_ceiling_1 = Triangle([p5, p6, p7, p5])
triangle_ceiling_2 = Triangle([p5, p7, p8, p5])

triangle_left_1 = Triangle([p1, p8, p4, p1])
triangle_left_2 = Triangle([p1, p5, p8, p1])

triangle_right_1 = Triangle([p2, p3, p7, p2])
triangle_right_2 = Triangle([p2, p7, p6, p2])

triangle_front_1 = Triangle([p1, p6, p5, p1])
triangle_front_2 = Triangle([p1, p2, p6, p1])

triangle_back_1 = Triangle([p4, p8, p7, p4])
triangle_back_2 = Triangle([p4, p7, p3, p4])


def border_to_stl(file: TextIO, border: Face):
    file.write("facet normal {x} {y} {z}\n".format(x=border.normal.x, y=border.normal.y, z=border.normal.z))
    file.write("outer loop\n")
    for vertex in border.points[0:-1]:
        file.write("vertex {x} {y} {z}\n".format(x=vertex.x, y=vertex.y, z=vertex.z))
    file.write("endloop\n")
    file.write("endfacet\n")


def to_stl(solid_name: str, borders: "list[Face]", export_name: str, path: str = None):
    filtered_borders = [border for border in borders if len(border.points) == 4]
    if path is None:
        current_path = Path.cwd()
        parent_directory = current_path.parent.absolute()
        path = parent_directory.joinpath("format", "export", "stl", "{file_name}.stl".format(file_name=export_name))
    file = open(path, "w")
    file.write("solid {solid_name}\n".format(solid_name=solid_name))
    for border in filtered_borders:
        border_to_stl(file, border)
    file.write("endsolid {solid_name}".format(solid_name=solid_name))


to_stl("test", [
    triangle_floor_1, triangle_floor_2,
    triangle_ceiling_1, triangle_ceiling_2,
    triangle_left_1, triangle_left_2,
    triangle_right_1, triangle_right_2,
    triangle_back_1, triangle_back_2,
    triangle_front_1, triangle_front_2
    ], "test_border")



from archaea.geometry.coordinate_array import CoordinateArray


# Dimensionless vector for integer and float values.
class Vector(CoordinateArray):
    coordinates: "list[(int,float)]"

    def __init__(self, *coordinates):
        filtered_coordinates = [entry for entry in coordinates if isinstance(entry, (int, float))]
        self.coordinates = filtered_coordinates
        super(CoordinateArray, self).__init__(filtered_coordinates)

    def is_parallel(self, other, tolerance: 1e-10):
        if self.dimension != other.dimension:
            return False

        length_product = self.magnitude() * other.magnitude()
        if length_product < tolerance:
            return True

        all([abs((value * other.magnitude()) - (other_value * self.magnitude()) < tolerance * length_product)
             for [value, other_value] in zip(self, other)])

    def is_perpendicular(self, other, tolerance: 1e-10):
        if self.dimension != other.dimension:
            return False

        return abs(self.dot(other)) < tolerance * self.magnitude() * other.magnitude()

    def normalize(self, tolerance=1e-10):
        if abs(self.magnitude() - 1) < tolerance:
            return self

        normalized_vector = self.scale(1.0 / self.magnitude())
        return self.__class__(*normalized_vector)

    def reverse(self):
        scaled_vector = self.scale(-1.0)
        return self.__class__(*scaled_vector)



from archaea.geometry.point3d import Point3d


class BoundingBox:
    min: Point3d
    max: Point3d
    center: Point3d
    area: float
    volume: float

    def __init__(self, minX: float, minY: float, minZ: float, maxX: float, maxY: float, maxZ: float) -> None:
        self.min = Point3d(minX, minY, minZ)
        self.max = Point3d(maxX, maxY, maxZ)
        self.__calculate_values()

    @classmethod
    def from_points(cls, points: "list[Point3d]"):
        x_values = [p.x for p in points]
        y_values = [p.y for p in points]
        z_values = [p.z for p in points]
        min_x = min(x_values)
        min_y = min(y_values)
        min_z = min(z_values)
        max_x = max(x_values)
        max_y = max(y_values)
        max_z = max(z_values)
        return cls(min_x, min_y, min_z, max_x, max_y, max_z)
    
    @classmethod
    def from_2_points(cls, p1: Point3d, p2: Point3d):
        return cls.from_points([p1, p2])

    def __calculate_values(self):
        self.center = Point3d(
            (self.max.x + self.min.x) / 2,
            (self.max.y + self.min.y) / 2,
            (self.max.z + self.min.z) / 2
        )
        x_dist = abs(self.max.x - self.min.x)
        y_dist = abs(self.max.y - self.min.y)
        z_dist = abs(self.max.z - self.min.z)
        self.volume = x_dist * y_dist * z_dist
        self.area = 2 * ((x_dist * y_dist) + (x_dist * z_dist) + (y_dist * z_dist))



from archaea.geometry.coordinate_array import CoordinateArray


# 2 dimensional vector.
class Vector2d(CoordinateArray):
    def __init__(self, *coordinates):
        filtered_coordinates = [entry for entry in coordinates if isinstance(entry, (int, float))]
        self.coordinates = filtered_coordinates[:2]
        super(CoordinateArray, self).__init__(filtered_coordinates)

    @property
    def x(self):
        return self.coordinates[0]

    @property
    def y(self):
        return self.coordinates[1]

    def outer_product(self, other):
        return (self.x * other.y) - (self.y * other.x)




import math

from archaea.geometry.coordinate_array import CoordinateArray
from archaea.geometry.vector import Vector


# 3 dimensional vector.
class Vector3d(Vector):
    def __init__(self, *coordinates):
        filtered_coordinates = [entry for entry in coordinates if isinstance(entry, (int, float))]
        self.coordinates = filtered_coordinates[:3]
        super(CoordinateArray, self).__init__(filtered_coordinates)

    @property
    def x(self):
        return self.coordinates[0]

    @property
    def y(self):
        return self.coordinates[1]

    @property
    def z(self):
        return self.coordinates[2]

    @property
    def length(self):
        return ((self.x ** 2) + (self.y ** 2) + (self.z ** 2)) ** 0.5

    def cross_product(self, other):
        return Vector3d(
            (self.y * other.z) - (self.z * other.y),
            (self.z * other.x) - (self.x * other.z),
            (self.x * other.y) - (self.y * other.x)
        )

    def azimuth_angle(self):
        result = 0
        if self.x > 0:
            result = (math.pi * 0.5) - math.atan(self.y / self.x)
        elif self.x < 0:
            result = (math.pi * 1.5) - math.atan(self.y / self.x)
        elif self.y > 0:
            result = 0
        elif self.y < 0:
            result = math.pi

        return math.degrees(result)

    def normalized(self):
        length = self.length
        if length == 0:
            return Vector3d(0, 0, 0)
        return Vector3d(self.x / length, self.y / length, self.z / length)

    def rotate(self, axis, angle, origin_point):
        # Calculate the vector from the origin to the point
        vector_to_point = self - origin_point.position_vector

        # Perform the rotation
        rotated_vector = vector_to_point.rotate(axis, angle)

        # Calculate the final position after rotation
        rotated_point = origin_point.position_vector + rotated_vector

        return Vector3d(rotated_point.x, rotated_point.y, rotated_point.z)




import functools
from archaea.geometry.vector3d import Vector3d
from archaea.geometry.loop import Loop
from archaea.earcut.earcut import earcut


class Face:
    outer_loop: Loop
    inner_loops: "list[Loop]"
    normal: Vector3d

    def __init__(self, outer_loop: Loop, inner_loops: "list[Loop]" = None):
        self.outer_loop = outer_loop
        self.normal = self.outer_loop.normal
        if inner_loops is None:
            self.inner_loops = []
        else:
            self.inner_loops = inner_loops

    @functools.cached_property
    def area(self):
        inner_loops_area = 0
        for loop in self.inner_loops:
            inner_loops_area += loop.area
        return self.outer_loop.area - inner_loops_area

    @functools.cached_property
    def plane(self):
        return self.outer_loop.plane()

    @functools.cached_property
    def has_hole(self):
        return self.inner_loops != 0

    def move(self, vector):
        moved_outer_loop = self.outer_loop.move(vector)
        moved_inner_loops = []
        for loop in self.inner_loops:
            moved_inner_loops.append(loop.move(vector))
        return self.__class__(moved_outer_loop, moved_inner_loops)

    def extrude(self, value):
        if value == 0:
            return self
        all_faces = [self]
        all_loops = [self.outer_loop]
        all_loops += self.inner_loops
        all_faces += self.outer_loop.extrude(value)
        for loop in self.inner_loops:
            loop_faces = loop.extrude(value, True)
            all_faces += loop_faces
        move_vector = Vector3d(*self.normal.scale(value))
        cap: Face = self.move(move_vector)
        all_faces.append(cap.reverse())
        return all_faces

    def reverse(self):
        reversed_outer_loop = self.outer_loop.reverse()
        reversed_inner_loops = [loop.reverse() for loop in self.inner_loops]
        return self.__class__(reversed_outer_loop, reversed_inner_loops)

    def offset(self, value):
        offset_outer_loop = self.outer_loop.offset(value)
        return self.__class__(offset_outer_loop, self.inner_loops)

    def mesh_polygon_vertices(self):
        all_loops = [self.outer_loop]
        all_loops += self.inner_loops
        all_points = self.outer_loop.points[:-1]
        all_uv_points = self.outer_loop.uv_points()
        hole_start_indexes = []
        for inner_loop in self.inner_loops:
            hole_start_indexes.append(len(all_uv_points))
            all_points += inner_loop.points[:-1]
            all_uv_points += inner_loop.uv_points(self.outer_loop.plane())

        hole_indices = None if len(hole_start_indexes) == 0 else hole_start_indexes
        flatten_coordinates = [num for sublist in all_uv_points for num in sublist]
        polygons = earcut(flatten_coordinates, hole_indices=hole_indices)

        n = 3
        triangular_polygons = [polygons[i:i + n] for i in range(0, len(polygons), n)]

        triangular_points = []
        for triangular_polygon in triangular_polygons:
            tp = triangular_polygon
            triangular_points.append([all_points[tp[0]], all_points[tp[1]], all_points[tp[2]]])

        return triangular_points



# Thanks for answer on the below link;
# https://stackoverflow.com/questions/12642256/find-area-of-polygon-from-xyz-coordinates

# determinant of matrix a
def det(a):
    return a[0][0] * a[1][1] * a[2][2] + a[0][1] * a[1][2] * a[2][0] + a[0][2] * a[1][0] * a[2][1] - a[0][2] * a[1][1] \
           * a[2][0] - a[0][1] * a[1][0] * a[2][2] - a[0][0] * a[1][2] * a[2][1]


# unit normal vector of plane defined by points a, b, and c
def unit_normal(a, b, c):
    x = det([[1, a[1], a[2]],
             [1, b[1], b[2]],
             [1, c[1], c[2]]])
    y = det([[a[0], 1, a[2]],
             [b[0], 1, b[2]],
             [c[0], 1, c[2]]])
    z = det([[a[0], a[1], 1],
             [b[0], b[1], 1],
             [c[0], c[1], 1]])
    magnitude = (x ** 2 + y ** 2 + z ** 2) ** .5
    if magnitude == 0:
        return 0, 0, 0
    return x / magnitude, y / magnitude, z / magnitude


# dot product of vectors a and b
def dot(a, b):
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


# cross product of vectors a and b
def cross(a, b):
    x = a[1] * b[2] - a[2] * b[1]
    y = a[2] * b[0] - a[0] * b[2]
    z = a[0] * b[1] - a[1] * b[0]
    return x, y, z


# area of polygon poly
def area(poly):
    if len(poly) < 3:  # not a plane - no area
        return 0

    total = [0, 0, 0]
    for i in range(len(poly)):
        vi1 = poly[i]
        if i is len(poly) - 1:
            vi2 = poly[0]
        else:
            vi2 = poly[i + 1]
        prod = cross(vi1, vi2)
        total[0] += prod[0]
        total[1] += prod[1]
        total[2] += prod[2]
    normal = unit_normal(poly[0], poly[1], poly[2])
    result = dot(total, normal)
    return round(result / 2, 5)



import math
from archaea.geometry.coordinate_array import CoordinateArray
from archaea.geometry.vector3d import Vector3d


# 3 dimensional point.
class Point3d(CoordinateArray):
    def __init__(self, *coordinates):
        filtered_coordinates = [entry for entry in coordinates if isinstance(entry, (int, float))]
        self.coordinates = filtered_coordinates[:3]
        super(CoordinateArray, self).__init__(filtered_coordinates)

    @property
    def x(self):
        return self.coordinates[0]

    @property
    def y(self):
        return self.coordinates[1]

    @property
    def z(self):
        return self.coordinates[2]

    def __sub__(self, other):
        return Vector3d(self.x - other.x, self.y - other.y, self.z - other.z)

    def __add__(self, vector):
        return Point3d(self.x + vector.x, self.y + vector.y, self.z + vector.z)

    def to_a(self):
        return [self.x, self.y, self.z]

    @property
    def position_vector(self):
        return Vector3d(self.x, self.y, self.z)

    def move(self, vector):
        return Point3d(self.x + vector.x, self.y + vector.y, self.z + vector.z)

    def vector_to(self, other):
        return Vector3d(other.x - self.x, other.y - self.y, other.z - self.z)

    def distance_to(self, other):
        return (((self.x - other.x) ** 2) + ((self.y - other.y) ** 2) + ((self.z - other.z) ** 2)) ** 0.5

    @classmethod
    def origin(cls):
        return cls(0, 0, 0)

    def rotate(self, axis, angle, origin=None):
        if origin is None:
            origin = Point3d(0, 0, 0)

        # Translate the point to the origin
        translated_point = self - origin

        # Create a rotation matrix
        axis = axis.normalized()
        c = math.cos(math.radians(angle))
        s = math.sin(math.radians(angle))
        t = 1 - c

        rotation_matrix = [
            [t * axis.x ** 2 + c, t * axis.x * axis.y - s * axis.z, t * axis.x * axis.z + s * axis.y],
            [t * axis.x * axis.y + s * axis.z, t * axis.y ** 2 + c, t * axis.y * axis.z - s * axis.x],
            [t * axis.x * axis.z - s * axis.y, t * axis.y * axis.z + s * axis.x, t * axis.z ** 2 + c]
        ]

        # Apply the rotation matrix to the translated point
        x_rotated = translated_point.x * rotation_matrix[0][0] + translated_point.y * rotation_matrix[1][
            0] + translated_point.z * rotation_matrix[2][0]
        y_rotated = translated_point.x * rotation_matrix[0][1] + translated_point.y * rotation_matrix[1][
            1] + translated_point.z * rotation_matrix[2][1]
        z_rotated = translated_point.x * rotation_matrix[0][2] + translated_point.y * rotation_matrix[1][
            2] + translated_point.z * rotation_matrix[2][2]

        # Translate the rotated point back to its original position
        rotated_point = Point3d(x_rotated, y_rotated, z_rotated) + origin

        return rotated_point
